#include <assert.h>
#include <errno.h>
#include <fstream.h>
#include <limits.h>
#include <string.h>

#include "stdprj.h"
#include "register.h"

// access to the one-dimensional array Data[k]
// as Data[i][j] with Data[NEntries][EntrySize]
int Register::Index(int i, int j)
{
   int k;

   // Exit on error!
   assert(i >= 0);
   assert(i < NEntries);
   assert(j >= 0);
   assert(j < EntrySize);

   // the following definition of k holds
   // 0 <= k <= (NEntries - 1) * EntrySize + (EntrySize - 1) =
   //         = NEntries * EntrySize - 1
   k = i * EntrySize + j;
   // Exit on error!
   assert(k >= 0);
   assert(k < (NEntries * EntrySize));

   return(k);
}

Register::Register(int es)
{
   int X;
   int Y;
   int Z;
   int i;

   // Exit on error!
   assert(es >= 1);

   // one char must have 8 bits
   assert(CHAR_BIT == 8);

   // set members of class to zero
   EntrySize = 0;
   NEntries = 0;
   EntryExists = FALSE;
   Mark = 0;

   // calculate necessary space:
   // X = es   is the space, that is necessary
   //          to save a state
   // Y = pow((UCHAR_MAX + 1), es) + 1 is the number of
   //          states, that must be taken up by the register
   // With X and Y the following C++-Code must run correctly
   // and without overflow:
   // Z = X * Y;
   // Data = new unsigned char [Z];

   // define X
   X = es;
   assert(X >= 1);

   // calculate Y = pow((UCHAR_MAX + 1), es) + 1
   Y = 1;
   // Check of overflow
   // Exit on error!
   assert(es < sizeof(int));
   for(i=0; i<es; i++)
   {
      Y *= UCHAR_MAX + 1;
   }
   // Exit on error!
   assert(Y >= 1);
   assert(Y < INT_MAX);
   Y++;

   // calculate Z = X * Y;
   Z = 0;
   for (i=0; i<X; i++)
   {
      // Check overflow
      // Exit on error!
      assert(Z <= (INT_MAX - Y));

      Z += Y;
   }
   // Exit on error!
   assert(Z >= 1);

   // allocate space from heap
   Data = new unsigned char [Z];
   // Exit on error!
   assert(Data);

   // remember size and number of entries in class
   EntrySize = X;
   NEntries = Y;
}

Register::~Register()
{
   // free allocated space
   delete [] Data;
}

int Register::GetEntrySize()
{
   return(EntrySize);
}

void Register::Reset()
{
   int k;

   // clear Register 
   for (k=0; k<NEntries*EntrySize; k++)
   {
      Data[k] = 0;
   }

   // clear state
   EntryExists = FALSE;
   Mark = 0;
}

int Register::Found()
{
   return(EntryExists);
}

void Register::Dump(const char * Path)
{
   int i, j;
   int k;
   int l;

   // Exit on error!
   assert(Path);
   l = strlen(Path);
   assert(errno == EZERO);
   assert(l > 0);

   // open file Path 
   ofstream F(Path);
   // Exit on error!
   assert(errno == EZERO);

   // save parameter of Register
   F << "parameter of Register:" << endl;
   // Exit on error!
   assert(errno == EZERO);

   F << "======================" << endl;
   // Exit on error!
   assert(errno == EZERO);

   F << "EntrySize       " << EntrySize << endl;
   // Exit on error!
   assert(errno == EZERO);

   F << "Mark            " << Mark << endl;
   // Exit on error!
   assert(errno == EZERO);

   F << "NEntries        " << NEntries << endl;
   // Exit on error!
   assert(errno == EZERO);

   F << "EntryExists     ";
   // Exit on error!
   assert(errno == EZERO);

   if (EntryExists)
   {
      F << "TRUE";
   }
   else
   {
      F << "FALSE";
   }
   // Exit on error!
   assert(errno == EZERO);

   F << endl << endl << endl << endl;
   // Exit on error!
   assert(errno == EZERO);

   // transfer data from Register to F
   F << "data of Register:" << endl;
   // Exit on error!
   assert(errno == EZERO);

   F << "=================" << endl;
   // Exit on error!
   assert(errno == EZERO);

   // seek position ind Data[]
   k = Index(0, 0);
   for (i=0; i<Mark; i++)
   {
      // finish dump
      if (i == (Mark - 1))
      {
         F << endl;
         // Exit on error!
         assert(errno == EZERO);
      }

      for (j=0; j<EntrySize; j++)
      {
         // formatted output
         if (j > 0)
         {
            F << SPACE << SPACE << SPACE;
            // Exit on error!
            assert(errno == EZERO);

            F << PERPENDICULAR;
            // Exit on error!
            assert(errno == EZERO);

            F << SPACE << SPACE << SPACE; 
            // Exit on error!
            assert(errno == EZERO);
         }

         F << "0x";
         // Exit on error!
         assert(errno == EZERO);

         F.width(2);
         // Exit on error!
         assert(errno == EZERO);

         F.fill('0');
         // Exit on error!
         assert(errno == EZERO);

         F.setf(F.uppercase);
         // Exit on error!
         assert(errno == EZERO);

         // dump Data[i][j]
         F << hex << (int) Data[k];
         // Exit on error!
         assert(errno == EZERO);

         // next Data[i][j]
         k++;

         // finish line
         if (j == EntrySize - 1)
         {
            F << endl;
            // Exit on error!
            assert(errno == EZERO);
         }
      }
   }
}

Register & Register::operator<<(Storage & S)
{
   int i, k;
   Storage Buffer(EntrySize);

   // Exit on error!
   assert(EntrySize == (S.GetSize()));

   // Exit, if there was a match before!
   assert(!EntryExists);

   // search in Register for content of S
   for (i=0; i<Mark; i++)
   {
      // seek position
      k = Index(i, 0);

      // access to "Data[i][0]"
      Buffer.FillFromArray(Data + k, EntrySize);

      // compare Buffer and S
      if (Buffer == S)
      {
         EntryExists = TRUE;
         break;
      }
   }

   // set k and Mark new
   k = Index(Mark, 0);
   // Exit on overflow!
   // (Actually, this cannot really happen!)
   assert(Mark < NEntries);
   Mark++;

   // enter data from S in "Data[k]"
   S.StoreToArray(Data + k, EntrySize);   

   return(*this);
}
